/*	AVRcam.cpp
*
*	Communication with AVRcam
*
*	#define these:
*		CAM_COMM_TYPE     class to use for serial output
*		CAM_COMM          variable name to use for serial output
*		CAM_TIMER         variable to use for (ms) timer
*
*	Diagnostic LEDs
*		#define the following (with names for LED outputs) if desired.
*		The corresponding event will toggle the LED.
*			CAM_LED_TX_ACK
*			CAM_LED_TX_TIMEOUT
*			CAM_LED_RX_TRACKING_DATA
*		Example:
*			#define CAM_LED_TX_TIMEOUT g_ledR
*
*	Revisions:
*		04-18-07	version 1 (cjs)
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

#include <avr/io.h>
#include <avr/pgmspace.h>
#include "Util.h"
#include "Debug.h"
#include "AVRcam.h"
#include "Timer.h"
#include "Comm.h"
#include "InOut.h"	// for (optional) LED

#ifndef CAM_TIMER
#error "CAM_TIMER must be defined"
#endif
#if !defined(CAM_COMM) || !defined(CAM_COMM_TYPE)
#error "CAM_COMM and CAM_COMM_TYPE must be defined"
#endif

extern TIMER CAM_TIMER;
extern CAM_COMM_TYPE CAM_COMM;
#ifdef CAM_LED_TX_ACK
extern OUT CAM_LED_TX_ACK;
#endif
#ifdef CAM_LED_TX_TIMEOUT
extern OUT CAM_LED_TX_TIMEOUT;
#endif
#ifdef CAM_LED_RX_TRACKING_DATA
extern OUT CAM_LED_RX_TRACKING_DATA;
#endif

/*
*	Object Tracking
*/

TOL AVRCAM::s_tol;
/* fTrackData is true when have data in tol, and it's in use by caller;
   fTrackData is set false when tol is available for new data */
bool AVRCAM::s_fTrackData = fFalse;

/*
*	Receive
*/

enum	// Rx State
{
	rsNil,	// nothing, yet
	rsA,
	rsC,
	rsK,
	rsTrackData,
	rsSkipTrack,	// getting tracking packet, but not recording it
};

static const uchar chStartTrack = 0x0a;
static const uchar chEndTrack = 0xff;
static const uchar chEndFrame = 0x0f;

static uchar s_rs = rsNil;
static uchar *s_pbRx = NULL;

/*
*	Transmit
*/

enum	// Transmit State
{
	tsSent,		// command sent, waiting for reply
	tsAck,
	tsNack,
	tsErr = tsNack,		// on error, resend as if nack
	tsIdle = tsNack,	// if get new Tx when idle, send from beginning, as if nack
};

/* (color map Tx should be done in < 15 ms) */
static const ulong msTimeoutTx = 300;

static uchar s_ts = tsIdle;


enum	// Transmit Command
{
	/* order is important! used in s_mptcpab and for sequences of commands */
	tcPing = 0,
	tcDisableTracking,
	tcSendColorMap,
	tcEnableTracking,
	tcMax
};

char s_mptcpab[tcMax][2] PROGMEM =	// NOTE: keep in sync with tc enum
{
	{ 'P', 'G' },
	{ 'D', 'T' },
	{ 'S', 'M' },
	{ 'E', 'T' },
};
char s_bEndCmd PROGMEM = '\r';

uchar AVRCAM::s_ctc;		// REVIEW: all s_ to class?
static char s_tc;
static const char *s_pabColorMap;
static char s_cbColorMap;


AVRCAM::AVRCAM()
{
	/* start out just sending ping to verify comms */
	s_tc = tcPing;
	s_ctc = 1;
	s_pabColorMap = NULL;
	s_cbColorMap = 0;
}

AVRCAM::AVRCAM(const char *pabColorMap, uint cb, bool fTrack)
{
	/* send ping, disable tracking, [send color map,] [enable tracking] */
	s_tc = tcPing;
	s_ctc = fTrack ? 4 : (pabColorMap == NULL ? 2 : 3);
	s_pabColorMap = pabColorMap;
	s_cbColorMap = cb;
}

void AVRCAM::DoComms() const
{
	Tx();
	if (FRx())
	{
//		Assert(s_ts == tsSent);
		if (s_ts == tsSent)
			s_ts = tsAck;
	}
}

void AVRCAM::Tx() const
{
	static ulong s_ms = 0;

	if (s_ctc == 0)		// nothing pending
		return;

	if (s_ts == tsAck)
	{
		/* last comm succeeded; keep going! */
#ifdef CAM_LED_TX_ACK
		CAM_LED_TX_ACK.Toggle();
#endif
LNext:
		if (--s_ctc == 0)	// no more commands
			s_ts = tsIdle;
		else
		{
			++s_tc;		// move to next command
			if (s_tc == tcSendColorMap && s_pabColorMap == NULL)
				goto LNext;
LSend:
			CAM_COMM.TxAbPgm(&s_mptcpab[(int)s_tc][0], 2);
			if (s_tc == tcSendColorMap)
				CAM_COMM.TxAbPgm(s_pabColorMap, s_cbColorMap);
			CAM_COMM.TxAbPgm(&s_bEndCmd, 1);
			s_ms = CAM_TIMER.MsCur();
			s_ts = tsSent;
		}
	}
	else if (s_ts == tsNack)
	{
		/* new command, or last command failed; send current command */
		goto LSend;
	}
	else
	{
		Assert(s_ts == tsSent);
		if (CAM_TIMER.FElapsed(s_ms, msTimeoutTx))
		{
#ifdef CAM_LED_TX_TIMEOUT
			CAM_LED_TX_TIMEOUT.Toggle();
#endif
			if (s_rs == rsSkipTrack)
				s_rs = rsNil;		// make sure Rx listening
			goto LSend;
		}
	}
}

/* returns true if ACK */
bool AVRCAM::FRx() const
{
	uchar ch;
	bool fDataOverRun;

	if (!CAM_COMM.FCharWaiting())
		return fFalse;

	ch = CAM_COMM.ChGet(&fDataOverRun);

	if (s_rs == rsNil)		// nothing, yet
	{
		if (ch == chStartTrack)		// 0x0A, cto, TO[], 0xFF
		{
			if (s_fTrackData)
			{
				/* buffer in use */
				s_rs = rsSkipTrack;
			}
			else
			{
				s_pbRx = (uchar *)&s_tol;
				s_rs = rsTrackData;
			}
		}
		else if (ch == 'A')
			s_rs = rsA;
		/* REVIEW: NCK? (nah... timeout takes care of this) */
	}
	else if (s_rs == rsTrackData)
	{
		/* reading tracking data */
		if (fDataOverRun)	// data lost...
		{
			/* need to keep reading until get chEndTrack */
			s_rs = rsSkipTrack;
		}
		else if (ch == chEndTrack)
		{
			s_rs = rsNil;
			if (s_pbRx == (uchar *)&s_tol.ato[s_tol.cto])
			{
				/* good data packet */
				s_fTrackData = fTrue;
#ifdef CAM_LED_RX_TRACKING_DATA
				CAM_LED_RX_TRACKING_DATA.Toggle();
#endif
			}
		}
		else if (s_pbRx < (uchar *)&s_tol + sizeof(TOL))
			*s_pbRx++ = ch;
		else	// hmmm... we've got too much for the buffer; looks like bad data
			s_rs = rsSkipTrack;
	}
	else if (s_rs == rsSkipTrack)
	{
		if (ch == chEndTrack)
			s_rs = rsNil;
	}
	else if (s_rs == rsA)
		s_rs = (ch == 'C') ? rsC : rsNil;
	else if (s_rs == rsC)
		s_rs = (ch == 'K') ? rsK : rsNil;
	else if (s_rs == rsK)
	{
		s_rs = rsNil;
		if (ch == '\r')		// ACK to current command!
			return fTrue;
	}

	return fFalse;
}

bool AVRCAM::FStartTrack(const char *pabColorMap, uint cb) const
{
	if (s_ctc > 0)	// still processing another request!
		return fFalse;
	if (pabColorMap == NULL)	// just turn on tracking
	{
		s_tc = tcEnableTracking;
		s_ctc = 1;
	}
	else
	{
		/* disable tracking, send new color map, enable tracking */
		s_pabColorMap = pabColorMap;
		s_cbColorMap = cb;
		s_tc = tcDisableTracking;
		s_ctc = 3;
	}
	return fTrue;
}

bool AVRCAM::FStopTrack() const
{
	if (s_ctc > 0)	// still processing another request!
		return fFalse;
	s_tc = tcDisableTracking;
	s_ctc = 1;
	return fTrue;
}
