/*	AVRcam.h
*
*	Communication with AVRcam
*
*	#define these:
*		CAM_COMM_TYPE     class to use for serial output
*		CAM_COMM          variable name to use for serial output
*		CAM_TIMER         variable to use for (ms) timer
*
*	Diagnostic LEDs
*		#define the following (with names for LED outputs) if desired.
*		The corresponding event will toggle the LED.
*			CAM_LED_TX_ACK
*			CAM_LED_TX_TIMEOUT
*			CAM_LED_RX_TRACKING_DATA
*		Example:
*			#define CAM_LED_TX_TIMEOUT g_ledR
*
*	Revisions:
*		04-18-07	version 1 (cjs)
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

#include <stdio.h>

/* (0,0) is upper left */
static const uint cxCam = 176;		// (0c-af observed)
static const uint cyCam = 144;		// (01-8f observed)

/*
*	Object Tracking
*/

typedef struct TO TO;
typedef struct TOL TOL;

struct TO	// Tracked Object
{
	uchar tc;				// tracked color (0-7)
	uchar xLeft, yTop;		// upper left ((0,0) at top left)
	uchar xRight, yBottom;	// lower right
};
static const int ctoMax = 8;	// max 8 tracked objects per frame

struct TOL	// TO List
{
	uchar cto;
	TO ato[ctoMax];
};


class AVRCAM
{
public:
	/* just initialize camera (ping to confirm communication) */
	AVRCAM();
	/* download the specified color map (NULL for no new map);
	   start tracking if fTrack (else leave tracking disabled) */
	AVRCAM(const char *pabColorMap, uint cb, bool fTrack);

	/* call DoComms every time through main program loop! */
	void DoComms() const;

	/* FTxIdle/Busy indicates whether the most recent transmission
	   is complete */
	inline bool FTxIdle() const { return s_ctc == 0; }
	inline bool FTxBusy() const { return s_ctc != 0; }

	/*
	*	Tracking
	*
	*	Once tracking is turned on, the tracking buffer will be filled;
	*	after a tracking packet is successful received, the buffer will
	*	be locked and new data will not be buffered until
	*	ReleaseTrackingBuffer is called.
	*/

	/* turns tracking on; if pabColorMap != NULL, will send that first;
	   color map must be in program memory; returns true if command
	   successfully queued (false returned if transmit is still working on
	   previous command sequence); note: this is safe to call even if tracking
	   is already on (e.g. to change the map) */
	bool FStartTrack(const char *pabColorMap, uint cb) const;

	/* turns tracking off; returns true if command successfully queued (false
	   returned if transmit is still working on previous command sequence) */
	bool FStopTrack() const;

	/* return buffer with tracking info if there's valid data in it,
	   else returns NULL */
	inline TOL *PtolCur() const { return s_fTrackData ? &s_tol : NULL; }

	/* call this to release the tracking buffer; this is necessary after each
	   successful tracking packet from the camera (i.e. when PtolCur returns !NULL);
	   note: this is safe to call even if buffer isn't in use */
	inline void ReleaseTrackingBuffer() const { s_fTrackData = fFalse; }

private:
	/* sends any pending transmissions */
	void Tx() const;
	/* checks for received characters; returns fTrue when gets ACK;
	   sets s_fTrackData when valid tracking packet received;
	   subsequent tracking packets will not be buffered until
	   s_fTrackData is cleared (e.g. via ReleaseTrackingBuffer) */
	bool FRx() const;

	/* static variables: one per class, not instance */
	static uchar s_ctc;
	static TOL s_tol;
	static bool s_fTrackData;
};
