/*	Comm.cpp
*
*	Serial communications.
*	- default communication at 8-N-1
*	- baud rate selected by #define (see below)
*	- (look at spec sheet for Rx/Tx pins, e.g. D0/1 on Mega16/32/644)
*
*	#define one of the following values to select the appropriate code
*	based on your microcontroller and desired communication rate.
*
*	define         USART   baud    ATmega...
*	------------   -----   ------   --------------------
*	COMM_9600       N/A     9600    8, 16, 32
*	COMM_0_9600      0      9600    644, 128
*	COMM_1_9600      1      9600    128
*	COMM_38k        N/A     38.4k   8, 16, 32
*	COMM_0_38k       0      38.4k   644, 128
*	COMM_1_38k       1      38.4k   128
*	COMM_0_76k       0      76.8k   644, 128
*	COMM_1_76k       1      76.8k   128
*	COMM_0_115k      0     115.2k   644, 128
*	COMM_1_115k      1     115.2k   128
*
*	Note, by default, transmit is buffered (so caller doesn't have to wait); to
*	disable that, define NO_COMM_BUFFER. (Disabling buffering will save 34 bytes SRAM.)
*
*	Note, by default, if there is a transmit buffer, and the buffer is full,
*	any call to a transmit method will wait for space to appear in the buffer
*	before stuffing another character into the buffer. If you want to make sure
*	the transmit calls never stall, and don't mind trashing your own output,
*	define NO_WAIT_COMM_BUFFER. If the buffer does get overwritten, an overflow
*	character (@) will be transmitted to indicate one or more characters got dropped.
*
*	Revisions:
*		08-20-06	buffered transmit, public TxCh
*		07-05-06	version 1 in master library
*		09-28-06	added (2) Pgm functions
*		10-07-06	20 MHz clock
*		10-20-06	optionally wait when comm buffer is full, bug fixes (TWS)
*		04-12-07	115.2k, 76.8k baud; TxAbPgm (cjs)
*		04-18-07	optional pfDataOverRun when calling ChGet (cjs)
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "Util.h"
#include "Debug.h"
#include "Comm.h"

#ifndef NO_COMM_BUFFER
static const int pwr = 5;				// 1 << 5 (= 2^5) = 32 chars in buffer
static const int cchBuf = 1 << pwr;
static const int mskBuf = cchBuf - 1;	// e.g. 00011111, ANDed with index to handle wrapping
static const char chOverflow = '@';

static char s_achBufComm[cchBuf];
volatile static char s_ichNext = 0;
volatile static char s_cchRemain = 0;
#endif

/* UBRR values...
*
*           ======= uC clk ========
*           8 MHz   16 MHz   20 MHz
*  BAUD:    -----   ------   ------
* 115.2k              16*      10   *U2X
*  76.8k              12       32*  *U2X
*  38.4k      12      25       32
*  19.2k      25      51       64
*  14.4k      34      68       86
*  9600       51     103      129
*  4800      103     207      259
*  2400      207     416      520
*/

#if defined(COMM_0_115k) || defined(COMM_1_115k)
#define BAUD_115k
#elif defined(COMM_0_76k) || defined(COMM_1_76k)
#define BAUD_76k
#elif defined(COMM_38k) || defined(COMM_0_38k) || defined(COMM_1_38k)
#define BAUD_38k
#elif defined(COMM_9600) || defined(COMM_0_9600) || defined(COMM_1_9600)
#define BAUD_9600
#else
#error "unrecognized COMM_* definition (see header comment)"
#endif // rate

#if defined(COMM_38k) || defined(COMM_9600)
#define USART_x
#elif defined(COMM_0_115k) || defined(COMM_0_76k) || defined(COMM_0_38k) || defined(COMM_0_9600)
#define USART_0
#elif defined(COMM_1_115k) || defined(COMM_1_76k) || defined(COMM_1_38k) || defined(COMM_1_9600)
#define USART_1
#else
#error "unrecognized COMM_* definition (see header comment)"
#endif // USART

#ifdef CLOCK_8M
	#ifdef BAUD_38k
		#define baudRegVal	12
	#elif defined BAUD_9600
		#define baudRegVal	51
	#endif // rate
#elif defined CLOCK_16M
	#ifdef BAUD_115k
		#define baudRegVal	16
		#define SET_U2X
	#elif defined BAUD_76k
		#define baudRegVal	12
	#elif defined BAUD_38k
		#define baudRegVal	25
	#elif defined BAUD_9600
		#define baudRegVal	103
	#endif // rate
#elif defined CLOCK_20M
	#ifdef BAUD_115k
		#define baudRegVal	10
	#elif defined BAUD_76k
		#define baudRegVal	32
		#define SET_U2X
	#elif defined BAUD_38k
		#define baudRegVal	32
	#elif defined BAUD_9600
		#define baudRegVal	129
	#endif // rate
#endif // CLOCK

#ifndef baudRegVal
#error "unsupported baud rate"
#endif

#ifdef USART_x
#define regUCSRB		UCSRB	// register with enable bits
#define bitRXEN			RXEN	// bit for enabling/disabling receive
#define bitTXEN			TXEN	// bit for enabling/disabling transmit
#define bitUDRIE		UDRIE	// bit for enabling data register empty interrupt
#define regUCSRC		UCSRC	// register for config info
#define regUBRRH		UBRRH	// register for high byte of baud value
#define regUBRRL		UBRRL	// register for high byte of baud value
#define regUDR			UDR		// register for data
#define regUCSRA		UCSRA	// register with status bits
#define bitUDRE			UDRE	// bit for whether data register is empty
#define bitRXC			RXC		// bit for whether receive is complete
#define bitDOR			DOR		// bit for whether there was a data overrun
#define vectUDRE		USART_UDRE_vect

#elif defined USART_0
#define regUCSRB		UCSR0B	// register with enable bits
#define bitRXEN			RXEN0	// bit for enabling/disabling receive
#define bitTXEN			TXEN0	// bit for enabling/disabling transmit
#define bitUDRIE		UDRIE0	// bit for enabling data register empty interrupt
#define regUCSRC		UCSR0C	// register for config info
#define regUBRRH		UBRR0H	// register for high byte of baud value
#define regUBRRL		UBRR0L	// register for high byte of baud value
#define regUDR			UDR0	// register for data
#define regUCSRA		UCSR0A	// register with status bits
#define bitUDRE			UDRE0	// bit for whether data register is empty
#define bitRXC			RXC0	// bit for whether receive is complete
#define bitDOR			DOR0	// bit for whether there was a data overrun
#if defined __AVR_ATmega168__
#define vectUDRE		USART_UDRE_vect
#else
#define vectUDRE		USART0_UDRE_vect
#endif

#elif defined USART_1
#define regUCSRB		UCSR1B	// register with enable bits
#define bitRXEN			RXEN1	// bit for enabling/disabling receive
#define bitTXEN			TXEN1	// bit for enabling/disabling transmit
#define bitUDRIE		UDRIE1	// bit for enabling data register empty interrupt
#define regUCSRC		UCSR1C	// register for config info
#define regUBRRH		UBRR1H	// register for high byte of baud value
#define regUBRRL		UBRR1L	// register for high byte of baud value
#define regUDR			UDR1	// register for data
#define regUCSRA		UCSR1A	// register with status bits
#define bitUDRE			UDRE1	// bit for whether data register is empty
#define bitRXC			RXC1	// bit for whether receive is complete
#define bitDOR			DOR1	// bit for whether there was a data overrun
#define vectUDRE		USART1_UDRE_vect

#else
#error "unknown USART"
#endif

char s_szCRLF[] PROGMEM = "\r\n";

bool COMM::s_fInit = fFalse;

COMM::COMM(bool fRx, bool fTx)
{
	if (!s_fInit)
		Init(fRx, fTx);
}

/* initializes serial transmit and receive */
void COMM::Init(bool fRx, bool fTx)
{
	s_fInit = fTrue;

	regUCSRC = 0x86; // 1000 0110 - asynch, no parity, 1 stop bit, 8 data bits
	regUBRRH = baudRegVal >> 8;
	regUBRRL = baudRegVal & 0xff;
#ifdef SET_U2X	// USART at 2x speed
	UCSR1A |= 0x02;		// double xmission clock
#endif

	if (fRx)
		SetBit(regUCSRB, bitRXEN);		// enable receive
	else
		ClearBit(regUCSRB, bitRXEN);	// disable receive

	if (fTx)
	{
		SetBit(regUCSRB, bitTXEN);		// enable transmit
#ifndef NO_COMM_BUFFER
#ifdef DEBUG
		/* fill the buffer with visible characters to detect bogus reads */
		for (uint ich = 0; ich < DIM(s_achBufComm); ++ich)
			s_achBufComm[ich] = '0' + ich;
#endif
		/* enable global interrupts; UDRIE will be enabled as appropriate below */
		sei();
#endif
	}
	else
		ClearBit(regUCSRB, bitTXEN);	// disable transmit
}

/* transmits text via serial connection */
void COMM::Print(const char *psz) const
{
	const char *pch;

	/* REVIEW: more efficient loading of buffer? */
	for (pch = psz; *pch != 0; ++pch)
		TxCh(*pch);
}

/* transmits text from program memory via serial connection */
void COMM::PrintPgm(const char *psz) const
{
	const char *pch;
	char ch;

	/* REVIEW: more efficient loading of buffer? */
	for (pch = psz; true; ++pch)
	{
		if ((ch = pgm_read_byte_near(pch)) == 0)
			break;
		TxCh(ch);
	}
}

/* transmits text plus end of line (e.g. CRLF) via serial connection */
void COMM::PrintLn(const char *psz) const
{
	if (psz != 0)
		Print(psz);
	PrintPgm(s_szCRLF);
}

/* transmits text from program memory plus end of line (e.g. CRLF) via serial connection */
void COMM::PrintPgmLn(const char *psz) const
{
	PrintPgm(psz);
	PrintPgm(s_szCRLF);
}
	
/* transmits an array of bytes stored in program memory */
void COMM::TxAbPgm(const char *pab, uint cb) const
{
	const char *pch;

	/* REVIEW: more efficient loading of buffer? */
	for (pch = pab; cb-- > 0; ++pch)
		TxCh(pgm_read_byte_near(pch));
}

/* transmits a single character (after waiting
   for any prev transmission to complete) */
void COMM::TxCh(char ch) const
{
#ifdef NO_COMM_BUFFER

	while (bit_is_clear(regUCSRA, bitUDRE))
		;
	regUDR = ch;

#else

#ifdef NO_WAIT_COMM_BUFFER
	if (s_cchRemain == DIM(s_achBufComm))
	{
		ClearBit(regUCSRB, bitUDRIE);
		
		s_achBufComm[(s_ichNext - 1) & mskBuf] = chOverflow;
		
		SetBit(regUCSRB, bitUDRIE);

		return;
	}
#else
	// wait until there is space in the buffer
	while (s_cchRemain == DIM(s_achBufComm))
		;
#endif

	{
		char cchRemain;

		/* disable (delay) interrupts so we have a consistent situation
		   with the comm output (we need to make sure it doesn't stop
		   when we expect it to be running) */
		ClearBit(regUCSRB, bitUDRIE);
		
		cchRemain = s_cchRemain;
		s_achBufComm[(s_ichNext + cchRemain) & mskBuf] = ch;
		s_cchRemain = cchRemain + 1;

		SetBit(regUCSRB, bitUDRIE);
	}

#endif
}

/* returns true if there is unread data in the receive buffer */
bool COMM::FCharWaiting() const
{
	return bit_is_set(regUCSRA, bitRXC);
}

/* returns the (next) character in the receive buffer (0 if none) */
char COMM::ChGet() const
{
	if (bit_is_clear(regUCSRA, bitRXC))
		return 0;
	return regUDR;
}

/* returns the (next) character in the receive buffer (0 if none) */
char COMM::ChGet(bool *pfDataOverRun) const
{
	*pfDataOverRun = bit_is_set(regUCSRA, bitDOR);
	if (bit_is_clear(regUCSRA, bitRXC))
		return 0;
	return regUDR;
}

#ifndef NO_COMM_BUFFER
/* this interrupt is called when the transmit buffer becomes empty
   (i.e. we can send another char) */
ISR(vectUDRE)
{
	char cchRemain = s_cchRemain;
	if (cchRemain == 0)
	{
		/* buffer is empty */
		ClearBit(regUCSRB, bitUDRIE);		// done for now; turn off interrupt
	}
	else
	{
		/* write next char */
		char ichNext = s_ichNext;
		regUDR = s_achBufComm[(int)ichNext];	// send next char (setting UDR clears UDRE)
#ifdef DEBUG
		s_achBufComm[(int)ichNext] = '#';
#endif
		s_ichNext = (ichNext + 1) & mskBuf;
		s_cchRemain = cchRemain - 1;
	}
}
#endif
