/*	Comm.h
*
*	Serial communications.
*	- default communication at 8-N-1
*	- baud rate selected by #define (see below)
*	- (look at spec sheet for Rx/Tx pins, e.g. D0/1 on Mega16/32/644)
*
*	#define one of the following values to select the appropriate code
*	based on your microcontroller and desired communication rate.
*
*	define         USART   baud    ATmega...
*	------------   -----   ------   --------------------
*	COMM_9600       N/A     9600    8, 16, 32
*	COMM_0_9600      0      9600    644, 128
*	COMM_1_9600      1      9600    128
*	COMM_38k        N/A     38.4k   8, 16, 32
*	COMM_0_38k       0      38.4k   644, 128
*	COMM_1_38k       1      38.4k   128
*	COMM_0_76k       0      76.8k   644, 128
*	COMM_1_76k       1      76.8k   128
*	COMM_0_115k      0     115.2k   644, 128
*	COMM_1_115k      1     115.2k   128
*
*	Note, by default, transmit is buffered (so caller doesn't have to wait); to
*	disable that, define NO_COMM_BUFFER. (Disabling buffering will save 34 bytes SRAM.)
*
*	Note, by default, if there is a transmit buffer, and the buffer is full,
*	any call to a transmit method will wait for space to appear in the buffer
*	before stuffing another character into the buffer. If you want to make sure
*	the transmit calls never stall, and don't mind trashing your own output,
*	define NO_WAIT_COMM_BUFFER. If the buffer does get overwritten, an overflow
*	character (@) will be transmitted to indicate one or more characters got dropped.
*
*	Revisions:
*		08-20-06	buffered transmit, public TxCh
*		07-05-06	version 1 in master library
*		09-28-06	added (2) Pgm functions
*		10-07-06	20 MHz clock
*		10-20-06	optionally wait when comm buffer is full, bug fixes (TWS)
*		04-12-07	115.2k, 76.8k baud; TxAbPgm (cjs)
*		04-18-07	optional pfDataOverRun when calling ChGet (cjs)
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

#pragma once

class COMM
{
public:
	/* constructor for setting up serial transmit and receive */
	COMM(bool fRx, bool fTx);

	/* transmits text via serial connection */
	void Print(const char *psz) const;
	void PrintPgm(const char *psz) const;	// text stored in program memory

	/* transmits text plus end of line (e.g. CRLF) via serial connection */
	void PrintLn(const char *psz = 0) const;
	void PrintPgmLn(const char *psz) const;	// text stored in program memory
	
	/* transmits an array of bytes stored in program memory */
	void TxAbPgm(const char *pab, uint cb) const;

	/* transmits a single character */
	void TxCh(char ch) const;

	/* returns true if there is unread data in the receive buffer */
	bool FCharWaiting() const;

	/* returns the (next) character in the receive buffer (0 if none) */
	char ChGet() const;
	/* returns the (next) character in the receive buffer (0 if none); also
	   sets/clears pfDataOverRun based on whether there was a receive buffer
	   was overrun */
	char ChGet(bool *pfDataOverRun) const;

private:
	static void Init(bool fRx, bool fTx);
	
	/* static variable: one per class, not instance */
	static bool s_fInit;
};
