/*	Servo.cpp
*
*	Servo control via PWM using Timer/Counter.
*
*	#define one of the following values to select the appropriate code
*	based on your microcontroller and desired Timer/Counter (T/C) usage.
*
*	define  T/C   ATmega...
*	------- ---   --------------------
*	SERVO_1  1    8, 16, 32, 644, 128
*	SERVO_3  3    128
*
*	Revisions:
*		07-05-06	version 1 in master library
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include "Util.h"
#include "Debug.h"
#include "Servo.h"

/* NOTE: also need to set usPerTick in Servo.h! */

#ifdef CLOCK_8M
/* clock = 8 Mhz / 64 = 125 kHz, 8 us/tick, count to 2500 for 20 ms */
static const uchar csServo = 0x03;		// 011: clk/64
static const uint ctick20ms = 2500;
/* clock = 8 Mhz / 8 = 1 MHz, 1 us/tick, count to 20000 for 20 ms */
//static const uchar csServo = 0x02;		// 010: clk/8
//static const uint ctick20ms = 20000;

#elif defined CLOCK_16M
/* clock = 16 Mhz / 64 = 250 kHz, 4 us/tick, count to 5000 for 20 ms */
static const uchar csServo = 0x03;		// 011: clk/64
static const uint ctick20ms = 5000;

#else
#error "unknown uC clock rate (see Servo.cpp, Util.h)"
#endif

#if !defined(__AVR_ATmega8__) && !defined(__AVR_ATmega16__) \
	&& !defined(__AVR_ATmega32__)&& !defined(__AVR_ATmega644__) \
	&& !defined(__AVR_ATmega128__)
#warning "unrecognized microcontroller (add/verify it, then update header comments in .cpp/.h)"
#endif

#ifdef SERVO_1
#define regTCNT		TCNT1
#define regTCCRnA	TCCR1A
#define regTCCRnB	TCCR1B
#define regICR		ICR1
#define regOCRnA	OCR1A
#define regOCRnB	OCR1B
#define regOCRnC	OCR1C

#elif defined SERVO_3
#define regTCNT		TCNT3
#define regTCCRnA	TCCR3A
#define regTCCRnB	TCCR3B
#define regICR		ICR3
#define regOCRnA	OCR3A
#define regOCRnB	OCR3B
#define regOCRnC	OCR3C

#else
#error "unrecognized SERVO_* definition (see header comment)"
#endif


/* values for setting COM bits in control register */
static const uint shComA = 6;	// bits to shift mskCom for COMnA bits (**00 0000)
static const uint shComB = 4;	// bits to shift mskCom for COMnB bits (00** 0000)
static const uint shComC = 2;	// bits to shift mskCom for COMnC bits (0000 **00)


/*
*	WGM = 1110	fast PWM, top=ICRn
*	COMx = 10	clear on match, set at top (fast PWM) [output on OCnx]
*/

char SERVO::s_fInit = fFalse;

SERVO::SERVO(char spc, ushort pwm)
{
	uchar shCom;

#ifdef __AVR_ATmega128__
	if (spc == spcC)
	{
		regOCRnC = 0;
		m_pregOCRnx = &regOCRnC;
		shCom = shComC;
	}
	else
#endif
	if (spc == spcA)
	{
		regOCRnA = 0;
		m_pregOCRnx = &regOCRnA;
		shCom = shComA;
	}
	else
	{
		Assert(spc == spcB);
		regOCRnB = 0;
		m_pregOCRnx = &regOCRnB;
		shCom = shComB;
	}

	if (!s_fInit)
		Init();

	/* set compare value */
	SetPWM(pwm);

	/* set COM bits for specified output */
	regTCCRnA |= 0x02 << shCom;
}

void SERVO::Init()
{
	uchar sregT;

	s_fInit = fTrue;

	/* need to disable interrupts when setting 2-byte values */
	sregT = SREG;		// save global interrupt flag
	cli();				// disable interrupts
	regTCNT = 0;
	regICR = ctick20ms;	// # ticks for 20ms (max value for counter before reset to 0)
	SREG = sregT;		// restore global interrupt flag

	regTCCRnA = 0x02;	// 00 00 00 10 (COMA, COMB, COMC, low WGM)
	regTCCRnB = 0x18 | csServo;	// 000 11 *** (misc, high WGM, CS)
//	regTCCRnC = 0x00;	// 000 00000 (FOC, unused)
}

void SERVO::SetPWM(ushort pwm) const
{
	uchar sregT;

	/* note that OCRnx is double-buffered -- it will be updated at TOP
	   (when TCNTn == ICRn) */

	/* need to disable interrupts when setting 2-byte values */
	sregT = SREG;		// save global interrupt flag
	cli();				// disable interrupts
	*m_pregOCRnx = pwm;	// (16-bit sfr: write high, then low -- compiler handles this!)
	SREG = sregT;		// restore global interrupt flag
}
