/*	main.cpp
*
*	Easter Egg Robot
*
*	- Drive forward; if see eggs with camera, steer toward them.
*	- When no more eggs visible, drive forward for a brief time,
*	  then turn (~3/4 rotation), looking for eggs; if none, just
*	  drive forward (and continue watching).
*	- If bump at any time, backup; when turn, watch for eggs and
*	  interrupt turn if see any (after min time)
*
*	DIP 1 on disables camera
*	DIP 4 on disables cone tracking
*
*	LED state at start, before bump to go:
*		camera enabled:
*			L blink: camera Tx timeout
*			L flicker: camera Rx tracking data
*			R on: Tx comms in progress
*			R on: bumper press registered
*		camera disabled:
*			R solid: camera disabled (DIP 1)
*			both on: bumper press registered
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

/*
new environment...
- download full color table
  (avrcam c'tor using color table, or directly to camera with avrisp)
- use Java app to watch tracking
- adjust map as appropriate
- confirm new map in cam EEPROM using avrdude & avrisp connected to camera
- set back to basic avrcam c'tor that just enables tracking
  (no color table download)
*/

#include <stdio.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "main.h"

/* optionally define one of these for LED activity when camera tracking */
//#define LED_SHOW_TURN
#define LED_SHOW_CONE_STATE


/*
*	Driving
*/

static const int msStartPause = 300; // delay before takes off (after bumper press)

static const int speedDrive = 250;
static const int speedBack = -500;
static const int speedTurn = 500;
static const int speedMin = 50;		// min speed used in DriveL/R

static const int pwmStop = PWMFromUs(1500);
static const int pwmPickup = PWMFromUs(2000);

enum		// Seek State
{
	ssFwd,		// driving forward, looking for eggs
	ssHaveEggs,	// have eggs in view
	ssSeekTurn,	// turning to look for eggs
	ssBumpMin,
	ssBackup = ssBumpMin,	// bump code has taken over
	ssBumpTurn,	// turning after bump & backup
};

static const int msPickup = 1200;		// time to continue forward after don't see any eggs (ssHaveEggs)
static const int msTurnSeek = 2800;		// time to turn when looking for eggs (ssSeekTurn)
static const int msBack = 800;			// time to back up (ssBackup)
static const int msTurnMin = 200;		// min time to turn before allow camera to take over (ssBumpTurn)
static const int msTurnLittle = 400;	// time to turn for a "small" turn (ssBumpTurn)
static const int msTurnLots = 1800;		// time to turn for a "large" turn (ssBumpTurn)
static const int msBumpClose = 5000;	// two bumps are close together if within 5 seconds

static const ulong msDriveMax = 30000;

static uchar s_ss = ssFwd;
static ulong s_msStart;		// set at beginning of ssFwd, ssHaveEggs, ssSeekTurn, ssBackup and ssBumpTurn

/*
*	Object Tracking
*/

/* (0,0) is upper left */
static const uchar dzTrackMin = 4;			// min size requirement for tracked object
//static const uchar dzTrackMinLong = 6;	// unused...
static const uchar yTrackMin = cyCam / 2;	// look for eggs in bottom of frame
static const uchar dxConeMin = 10;
static const uchar dyConeMin = 20;

enum	// Seeking for Cone
{
	/* note: order is assumed in code... */
	scDisabled,
	scWaiting,		// waiting to go into cone seeking mode
	scSeeking,		// looking for cone
	scFound,		// found cone
};

static const ulong msConeWait = 30000;	// time to collect eggs before search for cone
static const ulong msConeDrive = 20000;	// time to spend heading toward cone (after first find it)

static uchar s_sc = scWaiting;
static ulong s_msCone;		// when current cone state started (valid for scWaiting and scFound)

enum	// Tracked Colors
{
	tcPinkOrange = 0,
	tcUnused,
	tcYellow,
	tcGreen,
};

char s_abColorMapEggs[] PROGMEM =
{
	/* yeah, this is really ugly... see the XLS for something readable... */

//#define PINK_ORANGE_YELLOW_GREEN
//#define YELLOW_GREEN 
#define PINK_ORANGE_GREEN

#ifdef PINK_ORANGE_YELLOW_GREEN		// RTC
	' ', '0', ' ', '0', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '3', '2', ' ', '1', '6', '0', ' ', '1', '6', '0', ' ', '1', '6', '0', // R
	' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '4', '4', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '4', '8', ' ', '4', '8', ' ', '3', '2', ' ', '3', '2', ' ', '3', '2', // G
	' ', '1', '7', '6', ' ', '1', '7', '6', ' ', '1', '7', '6', ' ', '1', '7', '6', ' ', '1', '2', '8', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', // B
#endif
#ifdef PINK_ORANGE_YELLOW
	' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '3', '2', ' ', '1', '6', '0', ' ', '1', '6', '0', ' ', '1', '6', '0', // R
	' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '0', ' ', '0', ' ', '0', ' ', '3', '2', ' ', '3', '2', ' ', '3', '2', ' ', '3', '2', ' ', '3', '2', // G
	' ', '1', '6', '0', ' ', '1', '6', '0', ' ', '1', '6', '0', ' ', '1', '6', '0', ' ', '1', '2', '8', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', // B
#endif
#ifdef YELLOW_GREEN
	' ', '0', ' ', '0', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '3', '2', ' ', '3', '2', ' ', '3', '2', ' ', '3', '2', // R
	' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '4', '8', ' ', '4', '8', ' ', '3', '2', ' ', '3', '2', ' ', '3', '2', // G
	' ', '4', '8', ' ', '4', '8', ' ', '4', '8', ' ', '4', '8', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', // B
#endif
#ifdef PINK_ORANGE_GREEN
	' ', '0', ' ', '0', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', // R
	' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '2', '8', ' ', '1', '4', '4', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '1', '6', ' ', '0', ' ', '0', ' ', '0', // G
	' ', '1', '4', '4', ' ', '1', '4', '4', ' ', '1', '4', '4', ' ', '1', '4', '4', ' ', '1', '2', '8', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', // B
#endif
};
static const uint cbColorMapEggs = sizeof(s_abColorMapEggs);

char s_abColorMapCone[] PROGMEM =
{
	// OLD: tracked color 7 (0-based) -- R: 12-15, G: 0-4, B: 0-2
	// tracked color 7 (0-based) -- R: 14-15, G: 0-3, B: 0-2
	' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '1', ' ', '1', // R
	' ', '1', ' ', '1', ' ', '1', ' ', '1', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', // G
	' ', '1', ' ', '1', ' ', '1', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', ' ', '0', // B
};
static const uint cbColorMapCone = sizeof(s_abColorMapCone);

/* note: this is declared here instead of in globals.cpp because of color map */
AVRCAM g_cam(NULL, 0, fTrue);
//AVRCAM g_cam(s_abColorMapEggs, cbColorMapEggs, fTrue);
//AVRCAM g_cam(s_abColorMapCone, cbColorMapCone, fTrue);


void WaitForGo(bool fCamEnabled);
void CheckBump();
void CheckCamera();
bool FDriveFromCamera(TOL *ptol);
void DriveL(int speed);
void DriveR(int speed);


int main()
{
	bool fCamEnabled;

	fCamEnabled = !g_dip1.FClosed();
	if (g_dip4.FClosed())
		s_sc = scDisabled;

	if (!fCamEnabled)
		g_ledR.On();

	WaitForGo(fCamEnabled);
	g_timer.WaitMs(msStartPause);
	s_msCone = s_msStart = g_timer.MsCur();

	g_ledL.Off();
	g_ledR.Off();

	g_pickup.SetPWM(pwmPickup);
	DriveL(speedDrive);
	DriveR(speedDrive);

	while (1)
	{
		if (fCamEnabled)
			CheckCamera();
		CheckBump();
	}

	return 0;
}

/* wait for press, debounce, wait for release */
void WaitForGo(bool fCamEnabled)
{
	enum	// Wait State
	{
		wsWaitPress,
		wsDebounce,
		wsWaitRelease,
		wsDone,
	};

	uchar ws;
	ulong ms;

	ws = wsWaitPress;
	ms = 0;	// make compiler happy
	do
	{
		if (fCamEnabled)
		{
			g_cam.DoComms();
			g_ledR.Light(g_cam.FTxBusy());
			if (g_cam.PtolCur() != NULL)
			{
				g_ledL.Toggle();
				g_cam.ReleaseTrackingBuffer();
			}
		}

		if (ws == wsWaitPress)
		{
			if (g_bumpL.FClosed() || g_bumpR.FClosed())
			{
				g_ledL.On();
				g_ledR.On();
				ws = wsDebounce;
				ms = g_timer.MsCur();
			}
		}
		else if (ws == wsDebounce)
		{
			if (g_timer.FElapsed(ms, 5))
				ws = wsWaitRelease;
		}
		else if (ws == wsWaitRelease)
		{
			if (!g_bumpL.FClosed() && !g_bumpR.FClosed())
				ws = wsDone;
		}
	}
	while (ws != wsDone);
}

void CheckBump()
{
	enum	// Bump List
	{
		blNone = 0,			// nothing in current sequence
		fblLeft = 0x01,		// left (or sequence of L) most recent
		fblRight = 0x02,	// right (or sequance of R) most recent
		fblOpp = 0x04,		// bump prior to most recent was opposite of most recent
	};

	static uchar s_bl = blNone;
	static uchar s_fblCur;
	static ulong s_msLastBump;
	static ulong s_msWait;

	/* flush bump list if it's been long enough since last bump */
	if (s_bl != blNone && g_timer.FElapsed(s_msLastBump, msBumpClose))
		s_bl = blNone;

	if (g_bumpL.FClosed())
	{
		s_fblCur = fblLeft;
		goto LBackup;
	}
	else if (g_bumpR.FClosed())
	{
		s_fblCur = fblRight;
LBackup:
		s_ss = ssBackup;
		DriveL(speedBack);
		DriveR(speedBack);
		s_msStart = s_msLastBump = g_timer.MsCur();
	}
	else if (s_ss == ssBackup)
	{
		if (g_timer.FElapsed(s_msStart, msBack))
		{
			/* start turning */
			s_ss = ssBumpTurn;

			/* figure out whether to turn a little or a lot;
			   if LRL or RLR, probably trapped in corner, so turn a lot */
			/* e.g. new L bump:
			   prev   bl                   new state
			   ----   ------------------   ---------
			   X      0                    L
			   L      fblLeft              L
			   RL     fblLeft | fblOpp     L
			   R      fblRight             RL
			   LR     fblRight | fblOpp    X (and large turn)
			   */
			s_msWait = msTurnLittle;
			if (++s_bl >= 3)
			{
				s_msWait = msTurnLots;
				s_bl = 0;	// REVIEW: hacked use of s_bl...
			}
#ifdef OLD
			if (s_bl == 0 || s_bl & s_fblCur)	// new sequence, or same as last time
				s_bl = s_fblCur;	// (ensures that fblOpp isn't set)
			else if (s_bl & fblOpp)
			{
				/* 3 opposites in a row */
				s_msWait = msTurnLots;
				s_bl = blNone;
			}
			else	// opposite from last time
				s_bl = s_fblCur | fblOpp;
#endif

			if (s_fblCur & fblRight)		// obstacle on right
			{
				DriveR(speedTurn);
				DriveL(-speedTurn);
			}
			else
			{
				DriveL(speedTurn);
				DriveR(-speedTurn);
			}
			s_msStart = g_timer.MsCur();
		}
	}
	else if (s_ss == ssBumpTurn)
	{
		if (g_timer.FElapsed(s_msStart, s_msWait))
		{
			s_ss = ssFwd;
			s_msStart = g_timer.MsCur();
			DriveL(speedDrive);
			DriveR(speedDrive);
		}
	}

	/* REVIEW: need to handle "stuck" case when see eggs... */
	if (s_ss == ssFwd && g_timer.FElapsed(s_msStart, msDriveMax))
	{
		s_fblCur = fblRight;
		s_ss = ssBackup;
		DriveL(speedBack);
		DriveR(speedBack);
		s_msStart = s_msLastBump = g_timer.MsCur();
	}
}

void CheckCamera()
{
	TOL *ptol;
	bool fEggs;

	g_cam.DoComms();

	fEggs = fFalse;
	if ((ptol = g_cam.PtolCur()) != NULL)
	{
		/* we've got tracking data! let camera take over if we're
		   not handling bump */
		if (s_ss < ssBumpMin)
//			|| (s_ss == ssBumpTurn && g_timer.FElapsed(s_msStart, msTurnMin)) )
		{
			if (FDriveFromCamera(ptol))
				fEggs = fTrue;		// we found eggs and are tracking them
		}
		g_cam.ReleaseTrackingBuffer();
	}

	if (!fEggs)
	{
		/* didn't see any eggs this time through loop; possibilities:
		   - didn't get tracking data
		      - not time for new packet
			  - nothing visible
			  - lost comms with camera
		   - we didn't see anything interesting in tracking packet
		*/
		if (s_ss == ssHaveEggs)
		{
			/* keep doing last maneuver for a bit to wait for another
			   tracking packet, or to finish collecting if no more eggs visible */
			if (g_timer.FElapsed(s_msStart, msPickup))
			{
				/* spin and look for eggs */
				s_ss = ssSeekTurn;
				s_msStart = g_timer.MsCur();
				/* (turn in direction that doesn't require extra power boost) */
				DriveL(speedTurn);
				DriveR(-speedTurn);
			}
		}
		else if (s_ss == ssSeekTurn)
		{
			if (g_timer.FElapsed(s_msStart, msTurnSeek))
			{
				/* no eggs found during seeking turn; just wander... */
				s_ss = ssFwd;
				s_msStart = g_timer.MsCur();
				DriveL(speedDrive);
				DriveR(speedDrive);
			}
		}
	}
}

/* find all objects that look like eggs; calculate their center of mass
   and head for it; returns true if any eggs found */
bool FDriveFromCamera(TOL *ptol)
{
	int ito;
	uint dx, dy;
#ifdef OLD
	uint x;
	uchar cegg;
#else
	uchar weight;
	uint num, denom;	// numerator, denominator
#endif
	int speed;

	/* check/update cone seeking state */
	if (s_sc == scWaiting)
	{
		if (g_timer.FElapsed(s_msCone, msConeWait))
			s_sc = scSeeking;
	}
	else if (s_sc == scFound)
	{
		if (g_timer.FElapsed(s_msCone, msConeDrive))
		{
			/* we've spend long enough chasing; something's probably
			   not right (e.g. it's not really a cone, or it's someone
			   else's); give up for now */
			s_sc = scWaiting;
			s_msCone = g_timer.MsCur();
		}
	}

#ifdef OLD
	cegg = 0;
	x = 0;
#else
	num = denom = 0;
#endif
	for (ito = 0; ito < ptol->cto; ++ito)
	{
		dx = ptol->ato[ito].xRight - ptol->ato[ito].xLeft;
		dy = ptol->ato[ito].yBottom - ptol->ato[ito].yTop;
		/* look for cone starting in top half of frame
		   (after spend a while collecting eggs) */
		if (s_sc >= scSeeking && dy > dyConeMin && dx > dxConeMin && dy > dx
			&& ptol->ato[ito].tc == tcPinkOrange && ptol->ato[ito].yTop < cyCam/2)
		{
			/* pretend it's an egg; the rest of the code works fine! */
			/* we shouldn't find multiple of these, but if we
			   do, just head for one of them... */
			if (s_sc == scSeeking)
			{
				s_sc = scFound;
				s_msCone = g_timer.MsCur();
			}
#ifdef OLD
			cegg = 1;
			x = ptol->ato[ito].xRight + ptol->ato[ito].xLeft;
#else
			num = ptol->ato[ito].xRight + ptol->ato[ito].xLeft;
			denom = 1;
#endif
			goto LHaveEggs;
		}
		if (ptol->ato[ito].yTop > yTrackMin		// lower portion of camera's view
			&& dx >= dzTrackMin && dy >= dzTrackMin)
		{
#ifdef OLD
			++cegg;
			x += ptol->ato[ito].xRight + ptol->ato[ito].xLeft;
#else
			/* assign closer eggs (bottom of frame) higher weights */
			/* cyCam = 144, yTrackMin = 72, max yBottom = 143;
			   (143-72)/8+1 = 9, so weight range is 1-9 */
			weight = (ptol->ato[ito].yBottom - yTrackMin) / 8 + 1;
			num += (ptol->ato[ito].xRight + ptol->ato[ito].xLeft) * weight;
			denom += weight;
#endif
		}
	}

#ifdef OLD
	if (cegg == 0)
#else
	if (denom == 0)
#endif
		return fFalse;

LHaveEggs:
	s_ss = ssHaveEggs;
	/* record time when we (last) saw eggs */
	s_msStart = g_timer.MsCur();

#ifdef OLD
	x /= (2 * cegg);		// x is now center of mass
	Assert(x >= 0);
	Assert(x < cxCam);
	/* determine offset from center
	   (cxCam / 2 = 88, so offset from center max 88) */
	speed = (int)x - (int)(cxCam / 2);
#else
	num /= 2 * denom;		// num is now center of mass
	Assert(num < cxCam);
	/* determine offset from center
	   (cxCam / 2 = 88, so offset from center max 88) */
	speed = (int)num - (int)(cxCam / 2);
#endif
	speed *= 6;		// REVIEW: is 6 good?
	/* pos speed: would need to turn right to center eggs;
	   more speed to L motor */
	DriveL(speedDrive + speed);
	DriveR(speedDrive - speed);

#ifdef LED_SHOW_TURN
	g_ledL.Light(speed < 0);
	g_ledR.Light(speed > 0);
#elif defined(LED_SHOW_CONE_STATE)
	g_ledL.Light(s_sc == scSeeking);
	g_ledR.Light(s_sc == scFound);
#endif

	return fTrue;
}

/* L: - fwd, + rev; rev is weak dir */
/* speed -500 (full-speed reverse) to 500 (full-speed forward)
   (+/- 500 so speed corresponds to us...) */
void DriveL(int speed)
{
	if (speed < 0)
	{
		if (speed > -speedMin)	// make sure enough to move robot...
			speed = -speedMin;
		speed = speed * 6 / 5;	// need a little extra oomph for weak direction
		if (speed < -500)
			speed = -500;		// REVIEW: consts...
	}
	else if (speed < speedMin)
		speed = speedMin;
	else if (speed > 500)
		speed = 500;
	g_driveL.SetPWM(PWMFromUs(1500 - speed));
}

/* R: + fwd, - rev; fwd is weak dir */
void DriveR(int speed)
{
	if (speed > 0)
	{
		if (speed < speedMin)	// make sure enough to move robot...
			speed = speedMin;
		speed = speed * 6 / 5;	// need a little extra oomph for weak direction
		if (speed > 500)
			speed = 500;
	}
	else if (speed > -speedMin)
		speed = -speedMin;
	else if (speed < -500)
		speed = -500;		// REVIEW: consts...
	g_driveR.SetPWM(PWMFromUs(1500 + speed));
}
